 title lcdtst01
;
;DRTE      : February 2, 2001
;PURPOSE   : A programme to control an HD44780 LCD via a PIC 16F84
;AUTHOR    : Ian H.
;SCHEMATIC : LCDTST01.sch
;
;COMMENTS  : 
;
;PIC CONFIGURATION : 
;
;                            ------   ------
; DAT_C, 74C922 =>       RA2 |1    ---   18| RA1         <= DAT_B, 74C922
; DAT_D, 74C922 =>       RA3 |2          17| RA0         <= DAT_A, 74C922
;     BATT TEST => T0CKI/RA4 |3          16| OSC1/CLKIN  <= 2MHz XTAL
;        +5 VDC ->     /MCLR |4          15| OSC2/CLKOUT <= 2MHz XTAL
;           GND ->       Vss |5   16F84  14| Vdd         <- +5 VDC
;    74C922, DA  <-  INT/RB0 |6          13| RB7        ->  LCD EN (ENABLE)
;          LED1  <-      RB1 |7          12| RB6        ->  LCD CH (CHARACTER BITS)
;          LED2  <-      RB2 |8          11| RB5        ->  LCD CK (CLOCK)
;          LED3  <-      RB3 |9          10| RB4        ->  LED4
;                            ---------------
;REVISIONS : 

 LIST p=16F84, F=INHX8M, R=HEX
      errorlevel 0, -302, -305  
            ;302="register in operand not in "
            ;305="Using Default Destination"
 
 __CONFIG _CP_OFF & _WDT_OFF & _XT_OSC & _PWRTE_ON

 include "p16f84.inc"
 include "asc2lcd.inc"    ;The LCD alpha-numeric assignments. 
;
;CLOCK FREQ = 2.0 MHz

;*********************************************************
;VECTOR ASSIGNMENTS      
;*********************************************************
pic84    equ   0x04F      ;MAX memory address

;*********************************************************
;LCD pin assignments
;*********************************************************
EN       equ 7
CH       equ 6
CK       equ 5

RS       equ 3       ;Bit #4 in LCDFLAGS determines Command\Character for
                     ;the LCD
;*********************************************************
;FLAG REGISTER assignments
;*********************************************************
VT        equ  2      ;VALID TRANSMISSION  ("VT") from 74C922 to RB0

;*********************************************************
;MACROS
;*********************************************************
BANK_0 MACRO
 bcf  STATUS,RP0       ;RP0=STATUS.b5
 bcf  STATUS,RP1       ;RP1=STATUS.b6
 ENDM

BANK_1 MACRO
 bsf STATUS,RP0
 bcf STATUS,RP1
 ENDM

;*********************************************************
;REGISTER ASSIGNMENTS
;BANK #0
;*********************************************************
 CBLOCK 0x0C
w_                     ;Stores w during interrupts
STATUS_                ;A 'dummy' STATUS register
PCLATH_
FSR_
;LCD registers.
temp                   ;Holds HEX data prior to conversion to ASCII
w2                     ;Dummy w register used by nib_tx
count 
charac                 ;Data for the LCD
LCDflag                ;Flags R/W, cursor status etc  
d10000                 ;ASCII digits to be loaded into the LCD
d1000
d100
d10
d01
                       ;More counters .. used in the LCD routines
count_10               ;PAGE #1 count registers, available for
count_11               ;up to 3-loop counters
count_12  
count_13 
flags                  ;A 'decision' register
Keydata                ;Holds DATA NIBBLE from the 74C922
count_00               ;GP counters
count_01
count_02
count_03
 ENDC
 
;********************************************************
;Begin the MAIN PROGRAMME
;********************************************************
 PAGE
         org       0x00
start    goto      main

         org       0x04  
int      movwf     w_             ;Start interrupt routine
         movf      STATUS,0
         movwf     STATUS_ 
         movf      PCLATH,0       ;Save PCLATH
         movwf     PCLATH_
         movf      FSR,0
         movwf     FSR_
b1       call      kbdgrab        ;Grab the KPD data nibble.
         clrf      LCDflag
         bsf       LCDflag,RS     ;Set RS=1 for CHARACTER mode.  
         movf      Keydata,0
         movwf     temp           ;Now display the LCD readout
         call      hex2asc        ;Convert to ASCII
         movf      d01,0          ;And send to the LCD        
         call      nib_tx
         movf      d10,0
         call      nib_tx
         movf      d100,0
         call      nib_tx
                
         rlf       Keydata,1
         bcf       Keydata,0      ;Send a 0 to b0, an INPT?  ?????????????
         bCf       Keydata,EN     ;EN=b7. Make sure the LCD is disabled
         bcf       Keydata,CH     ;CH=b6
         bCf       Keydata,CK     ;CK=b5
         movf      Keydata,0
b2       movwf     PORTB          ;Send the result to the LEDs
               
int_ret  movlw     b'00011000'    ;Reset the INTCON bits BUT ...
         movwf     INTCON         ;.. let the PIC set GIE upon retfie
         movf      FSR_,0
         movwf     FSR
         movf      PCLATH_
         movwf     PCLATH
         movf      STATUS_,0      ;Return from interrupt
         movwf     STATUS
         swapf     w_             ;Flip & load the w register
         swapf     w_,w           ;without affecting the STATUS reg.
         retfie

;*************************************************************
;Initialise the PIC84's PORTS.
;
;PIN ASSIGNMENTS
;RB0  ..  INPUT. Reads DR  (Active High) from 74C922
;RB1  ..  LED1
;RB2  ..  LED2
;RB3  ..  LED3
;RB4  ..  LED4 
;RB5  ..  OUTPT to 74164 (CK)
;RB6  ..  OUTPT to 74164 (Data B = 'CH')
;RB7  ..  OUTPT to LCD (EN)
;
;RA0  ..  INPUT. Reads DAT_A from 74C922
;RA1  ..  INPUT. Reads DAT_B from 74C922
;RA2  ..  INPUT. Reads DAT_C from 74C922
;RA3  ..  INPUT. Reads DAT_D from 74C922
;RA4  ..  INPUT. NC. Use for battery test?
;
;*************************************************************
init     movlw   b'00011110'
         movwf   PORTB          ;Make PORTB LCD all LOW, avoid turn-on glitch??
         BANK_1
         movlw   b'00000001'    ;All except RB0 to be OUTPTs
         movwf   TRISB
         BANK_0     
         movlw   b'00011111'
         movwf   PORTA
         BANK_1
         movlw   0x01F
         movwf   TRISA
         BANK_0
         retlw   0x00

;*********************************************************************
;KEYPAD GRAB - responds to an interrupt from the DR pin on the 74C922, 
;grabs the Data nibble and stores it in KeyData. It then sets b3 in 
;flags before returning. 
;VT, the pin, must be same as VT in the flags register.
;*******************************************************************
kbdgrab movf     PORTA,0     ;Data are present.
        nop
        movwf    Keydata     ;Store the data
                             ;comf     Keydata,1   ;Invert the Active Low input.
        movlw    b'00001111' ;Make SURE the upper nibble
        andwf    Keydata,1   ;really is b'0000'.
        bsf      flags,VT    ;Tell the main programme
        nop                  ;Jump to here if no input
        retlw    0x00

;****************************************************************
;DLY160 ... a one-loop, 160 usec (2.0MHz xtal) delay routine
;Both LCD delay routines are called from PAGE #1 and return there.
;*****************************************************************
DLY160   movlw    0x05F     ;Calculate C1= 4F (d79) for T=160 us)
         movwf    count_11
         decfsz   count_11,1
          goto    $-1
         retlw    0x00

;****************************************************************
;DLY_5 ...  A two-loop, 5-m sec (2.0 MHz xtal) delay routine.
;Both LCD delay routines are called from PAGE #1 and return there.
;****************************************************************
DLY_5    movlw    0x02F       ;0x0F Calculate C2=$0A (d10) for T = 5 msec
         movwf    count_12
         movlw    0x0FF
         movwf    count_11
         decfsz   count_11,1
          goto    $-1
         decfsz   count_12,1
          goto    $-5
         retlw    0x00

;********************************************************************
;hex2asc ... a HEX to BCD & ASCII conversion routine. The hex number 
;to be converted must be in passed in 'temp'. 
;Experimental - gives correct results on MPLAB simulator.
;********************************************************************
hex2asc  clrf    count
         bcf     STATUS,C
         movlw   0x064
         subwf   temp,1
         btfss   STATUS,C
          goto   $+3
         incf    count,1
          goto   $-5
         movf    count,0       ;Retrieve the MSD
         movwf   d100
         movlw   0x064
         addwf   temp,1
         clrf    count
         bcf     STATUS,C
         movlw   0x0A
         subwf   temp,1
         btfss   STATUS,C      ;-ve?
          goto   $+3
         incf    count,1
         goto    $-5
         movf    count,0       ;Retrieve the middle digit
         movwf   d10
         movlw   0x0A
         addwf   temp,1        ;???
         movf    temp,0        ;Retrieve LSD
         movwf   d01
         movlw   0x030         ;Convert to ASCII
         addwf   d100,1
         addwf   d10,1
         addwf   d01,1
         clrf    count
         clrf    temp
         retlw   0x00

;*********************************************************
; LCD ROUTINES .. protocols derived from HD44780 data sheet
; LCDtx4  .. a routine to drive the HC44780 LCD in 4-bit 
;            mode via a 3-wire interface.
;
; Data byte must be passed in the 'charac' register 
; Data format ... D7 D6 D5 D4 RS x x x
; Registers/PORTs needed :-
;                     w
;                     charac
;                     count
;                     PORTB.CH =     (character\instrucs)
;                     PORTB.CK =     (SIPO clock)
;                     PORTB.EN =     (LCD Enable)
;
;This routine shifts a NIBBLE of Data into the LCD but does so 
;using a BYTE count. Data must be in the HIGH NIBBLE of the 
;charac register. LOW nibble bits are all zeroes.
;************************************************************
lcdtx4   movlw    0x08
         movwf    count
         bcf      STATUS,C
         bcf      PORTB,CK
         nop                    ;Rest between bit ops.
         bsf      PORTB,CH      ;Initialise the Data port.
rotate   rlf      charac,1      ;Rotate the the Data thru CARRY, MSB first.
         btfss    STATUS,C      ;Data bit = 1?
          bcf     PORTB,CH      ;No  - send a zero
         nop                    ;Rest between bit ops.
         bsf      PORTB,CK      ;Clock the bit thru
         nop 
         nop                    ;Rest between bit ops.
         bcf      PORTB,CK
         decfsz   count,1       ;Stepped down thru 8 bits?
          goto    rotate-1      ;No  - keep clockin'
         nop                    ;Rest between bit-ops.
         bsf      PORTB,EN      ;Enable the LCD and
         nop                    ;Rest between bit ops and
         nop                    ;transfer the Data.
         nop
         nop
         nop
         nop
         nop
         nop
         bcf      PORTB,EN
         nop
         bcf      PORTB,CH      ;Leave PORTB in a defined state. All LOW.
         retlw    0x00          ;Return to calling routine in PAGE #1

;**************************************************************
;NIB_TX takes a Data byte and puts the LS nibble into the HIGH
;nibble of the charac register. High nibble is DATA, low nibble bits
;are all zero. Data byte is passed in the w register.
;NIB_TX checks RS of the LCDflags register to determine CHARACTER or
;COMMAND mode for the LCD
;*****************************************************************
nib_tx  movwf    w2           ;w2 a VIRTUAL w register.. NOT w_ 
        bcf      PORTB,EN     ;Make sure the LCD starts off disabled
        movlw    b'11110000'  ;Isolate the upper nibble and
        andwf    w2,0         ;retrieve it in w.
        movwf    charac       ;Stash the processed byte in charac
        btfsc    LCDflag,RS   ;RS=0? COMMAND mode?
         bsf     charac,3     ;No - set RS line for CHARACTER mode.
        call     lcdtx4      ;and send it to the LCD.
        IFNDEF DEBUG
         call     DLY160      ;Wait 160 us till the LCD settles
        ENDIF
        swapf    w2,1         ;Invert the H & L nibbles
        movlw    b'11110000'
        andwf    w2,0         ;Isolate the new upper 4 digits 
        movwf    charac       ;and store in charac register.
        btfsc    LCDflag,RS   ;RS=0? COMMAND mode?
         bsf     charac,3     ;No  - set RS line for CHARACTER mode.
        call     lcdtx4      ;and send it to the LCD.
        IFNDEF DEBUG
         call     DLY160      ;Wait 160 uS till the LCD settles.
        ENDIF
        retlw    0x00         ;Go back to page #1 calling routine

;**********************************************************
;Main Programme  .. begin from upper half of Bank #0. Turn 
;the interrupts OFF till we're configured.
;**********************************************************
main    bcf      INTCON,GIE   ;Switch off interrupts while we  ...  
        BANK_0  
        call     init         ;Initialise the ports
         
;***********************************************************
;Power up and configure the LCD
;***********************************************************
pwr_up   clrf     LCDflag
         bcf      LCDflag,RS  ;Clear RS.. COMMAND mode ENABLED
         IFNDEF DEBUG
          call    DLY_5       ;Wait 15 ms after switch-on
          call    DLY_5
          call    DLY_5
          call    DLY_5
         ENDIF

         movlw    b'00110000' ;Write 0x03 to the LCD three
         movwf    charac      ;times with suitable delays
         call     lcdtx4      ;RS=0 for COMMAND instructions
         IFNDEF DEBUG
          call    DLY_5
         ENDIF

         movlw    b'00110000'
         movwf    charac
         call     lcdtx4
         IFNDEF DEBUG
          call    DLY160
         ENDIF

         movlw    b'00110000'
         movwf    charac
         call     lcdtx4
         IFNDEF DEBUG
          call    DLY160
         ENDIF

         movlw    b'00100000'  ;4 bit interface     
         movwf    charac
         call     lcdtx4
         IFNDEF DEBUG
          call    DLY160        
         ENDIF

;***************************************************************
;The following instrucs require 2 only NIBBLE WRITES to the LCD.
;Pass the parameters to the NIB_TX subroutine in w. NIB_TX will
;call the lcd_tx4 subroutine. BEWARE the nest level!!
;***************************************************************
         movlw    b'00101000'   ;Function select
         call     nib_tx
         IFNDEF DEBUG
          call    DLY160
         ENDIF

         movlw    b'00001110'   ;Write ON/OFF
         call     nib_tx
         IFNDEF DEBUG
          call    DLY160
         ENDIF

         movlw    b'00000110'   ;Write ENT mode
         call     nib_tx
         IFNDEF DEBUG
          call    DLY160
         ENDIF

         movlw    b'00000011'   ;Return CURSOR & LCD to home
         call     nib_tx
         IFNDEF DEBUG
          call    DLY160
         ENDIF
                      
         IFNDEF DEBUG
          call    DLY_5          ;Let things settle before proceeding
         ENDIF

setint  nop   
        BANK_1
        movlw    0x0D
        movwf    OPTION_REG
        BANK_0
        clrf     TMR0
        movlw    b'10011000'      ;Enable the interrupts
        movwf    INTCON
        clrf     flags            ;Initialise the flags

;************************************************************
;Then display a "WAKE UP" message ...
;************************************************************
wakeup   nop
         clrf     LCDflag
         bsf      LCDflag,RS    ;Set RS=1 for CHARACTER mode.  
         movlw    lcd_I         ;Load 'I'
         call     nib_tx
         IFNDEF DEBUG
          call    DLY160
         ENDIF
         movlw    lcd_A         ;Load 'A'
         call     nib_tx
         IFNDEF DEBUG
          call    DLY160
         ENDIF
         movlw    lcd_N         ;Load 'N'
         call     nib_tx
         IFNDEF DEBUG
          call    DLY160
         ENDIF
         movlw    b'00010010'    ;LED pattern should be OFF/ON/ON/OFF
         movwf    PORTB          ;No stimulus to the interrupt pin
                 
fin      goto    $                 ;Hold it here
         end
